unit hiCookieManager;

interface

uses Windows, Share, Kol, WinInet, uri;

type
  
  TCookie = class
    private
      function GetEffectiveDomain: string;
      function GetEffectivePath: string;

      function GetCookie: string;
      function GetServerCookie: string;
      function GetClientCookie: string;
      
    public
      ExpiresDate: TDateTime;
      
      Domain: string;
      Expires: string;
      Name: string;
      Path: string;
      Value: string;
      
      URLHost: string;
      URLPath: string;
      URLProto: string;
      
      property ClientCookie: string read GetClientCookie;
      property CookieText: string read GetCookie;
      property EffectiveDomain: string read GetEffectiveDomain;
      property EffectivePath: string read GetEffectivePath;
      property ServerCookie: string read GetServerCookie;

      
      function IsValidCookie(Url: string): Boolean; overload; // ,        AServerHost
      function IsValidCookie(Url: TURI): Boolean; overload; //
      function SetCookie(Url, CookieStr: string): Boolean; overload; //  ,    (  'Set-Cookie')
      function SetCookie(Url: TURI; CookieStr: string): Boolean; overload; //
      function IsExpired(Dt: TDateTime): Boolean; // ,      . Dt -   
      function TheSame(C: TCookie): Boolean; //    C   
      
      function AsString: string; //      
      function GetStringForSave: string; //      
      function SetCookieFromSaveStr(S: string): Boolean; //     
  end;
  
  ThiCookieManager = class
    private
      FCookieList: PList;
      FArr: PArray;
      FResult: string;
      FCSList: TRTLCriticalSection;
      
      function _Get(var Item:TData; var Val:TData):boolean;
      function _Count:Integer;
      procedure FreeObjects;
      procedure AddToList (C: TCookie);
      procedure InternalAddCookies(URL, Cookies: string; Source: Integer); 
      function InternalGetCookies(URL: string; Source: Integer): string;
      
    public
    
      _prop_Filename: string;
      _prop_Name: string;
      _prop_Source: Integer;
      
      _data_Text:THI_Event;
      _data_URL:THI_Event;
      _data_Index:THI_Event;
      _data_Filename:THI_Event;
      
      _event_onChange:THI_Event;
      _event_onGetCookies:THI_Event;
      _event_onGetIECookies:THI_Event;

      procedure _work_doAddCookies(var _Data:TData; Index:word);
      procedure _work_doAddIECookies(var _Data:TData; Index:word);
      procedure _work_doGetCookies(var _Data:TData; Index:word);
      procedure _work_doGetIECookies(var _Data:TData; Index:word);
      procedure _work_doSave(var _Data:TData; Index:word);
      procedure _work_doLoad(var _Data:TData; Index:word);
      procedure _work_doValue(var _Data:TData; Index:word);
      procedure _work_doDelete(var _Data:TData; Index:word);
      procedure _work_doClear(var _Data:TData; Index:word);
      
      procedure _var_Cookies(var _Data:TData; Index:word);
      procedure _var_List(var _Data:TData; Index:word);
      
      constructor Create;
      destructor Destroy; override;
      
      procedure AddCookies(URL, Cookies: string); 
      function GetCookies(URL: string): string;
      function GetInterfaceCookieMan: ThiCookieManager;   
  end;
  
  function IsDomainAcceptable(Host, Domain: string): Boolean;
  function IsPathMatchCookiePath(TargetPath, CookiePath: string): Boolean;
  function IsHostDomainMatches(Host, Domain: string): Boolean;
  
  function InternetTimeToSystemTime(lpszTime: PChar; pST: PSystemTime; dwReserved: DWORD): BOOL; stdcall; external 'wininet.dll' name 'InternetTimeToSystemTime';

var
  OffsetFromUTC: TDateTime;
  
implementation


function GetOffsetFromUTC: TDateTime;
var
  iBias: Integer;
  tmez: TTimeZoneInformation;
  ST: TSystemTime;
begin
  iBias := 0;
  FillChar(ST, SizeOf(TSystemTime), #0);
  case GetTimeZoneInformation(tmez) of
    TIME_ZONE_ID_UNKNOWN  :
       iBias := tmez.Bias;
    TIME_ZONE_ID_DAYLIGHT :
    begin
      iBias := tmez.Bias;
      if tmez.DaylightDate.wMonth <> 0 then
      begin
        iBias := iBias + tmez.DaylightBias;
      end;
    end;
    TIME_ZONE_ID_STANDARD :
    begin
      iBias := tmez.Bias;
      if tmez.StandardDate.wMonth <> 0 then
      begin
        iBias := iBias + tmez.StandardBias;
      end;
    end
  end;
  ST.wHour := Abs(iBias) div 60;
  ST.wMinute := Abs(iBias) mod 60;
  SystemTime2DateTime(ST, Result);
  
  {The GetTimeZone function returns values oriented towards converting
   a GMT time into a local time.  We wish to do the opposite by returning
   the difference between the local time and GMT.  So I just make a positive
   value negative and leave a negative value as positive}
  if iBias > 0 then Result := 0.0 - Result;
end;


function AddCookieProperty(AProperty, AValue, ACookie: string): string;
begin
  if Length(AValue) > 0 then
  begin
    if Length(ACookie) > 0 then
    begin
      ACookie := ACookie + '; ';
    end;
    ACookie := ACookie + AProperty + '=' + AValue;
  end;

  Result := ACookie;
end;

function CookieStrToDateTime(AString: string): TDateTime;
// Wed, 01-May-2013 15:34:00 GMT
var
  ST: TSystemTime;
begin
  FillChar(ST, SizeOf(TSystemTime), 0);
  InternetTimeToSystemTime(PChar(AString), @ST, 0);
  SystemTime2DateTime(ST, Result);
end;

{
When searching the cookie list for valid cookies, a comparison of the domain attributes of the cookie is made with the Internet domain name of the host from which the URL will be fetched. If there is a tail match, then the cookie will go through path matching to see if it should be sent. "Tail matching" means that domain attribute is matched against the tail of the fully qualified domain name of the host. A domain attribute of "acme.com" would match host names "anvil.acme.com" as well as "shipping.crate.acme.com".
}

{
RFC 6265
   The Domain attribute specifies those hosts to which the cookie will
   be sent.  For example, if the value of the Domain attribute is
   "example.com", the user agent will include the cookie in the Cookie
   header when making HTTP requests to example.com, www.example.com, and
   www.corp.example.com.  (Note that a leading %x2E ("."), if present,
   is ignored even though that character is not permitted, but a
   trailing %x2E ("."), if present, will cause the user agent to ignore
   the attribute.)  If the server omits the Domain attribute, the user
   agent will return the cookie only to the origin server.
   
5.1.3.  Domain Matching

   A string domain-matches a given domain string if at least one of the
   following conditions hold:

   o  The domain string and the string are identical.  (Note that both
      the domain string and the string will have been canonicalized to
      lower case at this point.)

   o  All of the following conditions hold:

      *  The domain string is a suffix of the string.

      *  The last character of the string that is not included in the
         domain string is a %x2E (".") character.

      *  The string is a host name (i.e., not an IP address)
}

// Firefox, Opera, Chrome:
//       -       ,   
// IE
//      ,      .

//             
//         .

function IsHostDomainMatches(Host, Domain: string): Boolean; // ,   Host  Domain
var
  LH, LD: Integer;
begin
  Result := False;
  
  LH := Length(Host);
  LD := Length(Domain);
  
  if (LH = 0) or (LD = 0) then Exit;
  
  
  // Host:   a.b.c.com
  // Matches domain: a.b.c.com, .b.c.com, .c.com
  // Not matches domain: b.c.com, c.com (no leading dot)
  
  // Host: y.com
  // Domain: y.com .y.com
  if (Host = Domain) or ('.' + Host = Domain) then
  begin
    Result := True;
    Exit;
  end;
  
  if Domain[1] <> '.' then Exit;
  
  if Copy(Host, LH-LD+1, LD) <> Domain then Exit;

  Result := True;
end;


function IsPathMatchCookiePath(TargetPath, CookiePath: string): Boolean; // ,     CookiePath   TargetPath
var
  LT, LC: Integer;
begin
  Result := False;
  
  LT := Length(TargetPath);
  LC := Length(CookiePath);
  
  if LT = 0 then Exit;
  
  if LC = 0 then
  begin
    CookiePath := '/';
    LC := 1;
  end;
  
  if TargetPath[1] <> '/' then
  begin
    TargetPath := '/' + TargetPath;
    Inc(LT);
  end;
  
  if TargetPath[LT] <> '/' then
  begin
    TargetPath := TargetPath + '/';
    //Inc(LT);
  end;
  
  if CookiePath[1] <> '/' then
  begin
    CookiePath := '/' + CookiePath;
    Inc(LC);
  end;
  
  if CookiePath[LC] <> '/' then
  begin
    CookiePath := CookiePath + '/';
    Inc(LC);
  end;
  
  // CookiePath: /aaa/, /bbb/
  // TargetPath: /aaa/, /bbb/ccc/
  if Copy(TargetPath, 1, LC) = CookiePath then Result := True;

end;




function PosEx(const SrcStr: string; SearchChar: Char; StartPos: Integer = 1): Integer;
begin
  if StartPos < 1 then StartPos := 1;
  for Result := StartPos to Length(SrcStr) do
  begin
    if SrcStr[Result] = SearchChar then Exit;
  end;
  Result := 0;
end;



// ,     Domain,   Host
{
Only hosts within the specified domain can set a cookie for a domain and domains must have at least two (2) or 
three (3) periods in them to prevent domains of the form: ".com", ".edu", and "va.us". Any domain that 
fails within one of the seven special top level domains listed below only require two periods. 
Any other domain requires at least three. 
The seven special top level domains are: "COM", "EDU", "NET", "ORG", "GOV", "MIL", and "INT". 

The default value of domain is the host name of the server which generated the cookie response.
}

{
RFC 6265

   The user agent will reject cookies unless the Domain attribute
   specifies a scope for the cookie that would include the origin
   server.  For example, the user agent will accept a cookie with a
   Domain attribute of "example.com" or of "foo.example.com" from
   foo.example.com, but the user agent will not accept a cookie with a
   Domain attribute of "bar.example.com" or of "baz.foo.example.com".

   NOTE: For security reasons, many user agents are configured to reject
   Domain attributes that correspond to "public suffixes".  For example,
   some user agents will reject Domain attributes of "com" or "co.uk".
   (See Section 5.3 for more information.)
   

6.   If the domain-attribute is non-empty:

           If the canonicalized request-host does not domain-match the
           domain-attribute:

              Ignore the cookie entirely and abort these steps.

           Otherwise:

              Set the cookie's host-only-flag to false.

              Set the cookie's domain to the domain-attribute.
   
}

function IsDomainAcceptable(Host, Domain: string): Boolean; 
var
  LH, LD: Integer;
begin
  Result := False;

  LH := Length(Host);
  LD := Length(Domain);
  
  if (LH = 0) or (LD = 0) then Exit;
  
  // Matches:
  
  // Host:   y.com, x.y.com
  // Domain: y.com,  .y.com

  // Host: y.com
  // Domain: y.com, .y.com
  if (Host = Domain) or ('.' + Host = Domain) then
  begin
    Result := True;
    Exit;
  end;
  
  if Domain[1] <> '.' then
  begin
    Domain := '.' + Domain;
    Inc(LD);
  end;
  
  
  // Domain: .com (1 dot)
  if PosEx(Domain, '.', 2) = 0 then Exit;
  
  // Host: x.y.com
  // Domain: .a.y.com
  if LD >= LH then Exit;
  
  // Host: a.x.y.com
  // Domain: .y.com (a.x contains a dot - RFC 2965)
  if PosEx(Copy(Host, 1, LH-LD), '.') <> 0 then Exit;
  
  // Host: x.y.com
  // Domain: .y.com
  if Copy(Host, LH-LD+1, LD) = Domain then Result := True;

end;



// ========== TCookie ========== //

function TCookie.GetEffectiveDomain: string;
begin
  if Length(Domain) > 0 then Result := Domain else Result := URLHost;
end;

function TCookie.GetEffectivePath: string;
begin
  if Length(Path) > 0 then Result := Path else Result := '/';
end;


function TCookie.GetCookie: string;
begin
  Result := AddCookieProperty(Name, Value, '');
  Result := AddCookieProperty('path', Path, Result);
  Result := AddCookieProperty('expires', Expires, Result);
  Result := AddCookieProperty('domain', Domain, Result);
end;

function TCookie.GetServerCookie: string;
begin
  Result := GetCookie;
end;


function TCookie.GetClientCookie: string;
begin
  Result := Name + '=' + Value;
end;


function TCookie.IsValidCookie(Url: string): Boolean;
var
  U: TURI;
begin
  U := TURI.Create(Url);
  if Length(U.Protocol) = 0 then U.Protocol := 'http';
  Result := IsValidCookie(U);
  U.Free;
end;

function TCookie.IsValidCookie(Url: TURI): Boolean;
begin
  Result := False;
  
  if Length(Url.Host) = 0 then Exit;
  
  //      Domain -   domain-match
  if Length(Domain) > 0 then
  begin
    if not IsHostDomainMatches(Url.Host, Domain) then Exit;
  end
  else //        
    if Url.Host <> URLHost then Exit;
  
  //      Path -   path-match
  if Length(Path) > 0 then
  begin
    if not IsPathMatchCookiePath(Url.Path, Path) then Exit;
  end;
  
  //  
  //   URL   ,        
  //  ,   ,   http
  if URLProto <> Url.Protocol then Exit;

  Result := True;
end;


function TCookie.SetCookie(Url, CookieStr: string): Boolean;
var
  U: TURI;
begin
  U := TURI.Create(Url);
  if Length(U.Protocol) = 0 then U.Protocol := 'http';
  Result := SetCookie(U, CookieStr);
  U.Free;
end;

function TCookie.SetCookie(Url: TURI; CookieStr: string): Boolean;
var
  s, s2: string;
  ParamList: PStrList;
begin
  URLHost := Url.Host;
  URLPath := Url.Path;
  URLProto := Url.Protocol;
  
  Domain := '';
  Expires := '';
  ExpiresDate := 0;
  Name := '';
  Path := '';
  Value := '';
  
  Result := False;
  
  
  
  if (Length(URLHost) = 0) or (Length(CookieStr) = 0) then Exit;

  
  ParamList := NewStrList;

  while Length(CookieStr) > 0 do
  begin
    s := Parse(CookieStr, ';');
    ParamList.Add(Trim(s));
  end;
  
  s2 := ParamList.Items[0]; // Cookie value
  s := Parse(s2, '=');
  Name := s;
  Value := s2;
  ParamList.Delete(0); //        ,    
  
  Domain := LowerCase(ParamList.Values['domain']);
  if Length(Domain) > 1 then
  begin
    {RFC 2965
      Domain=value
      OPTIONAL.  The value of the Domain attribute specifies the domain
      for which the cookie is valid.  If an explicitly specified value
      does not start with a dot, the user agent supplies a leading dot.
    }
    if Domain[1] <> '.' then Domain := '.' + Domain;
    
    {
      RFC 6265
      If the first character of the attribute-value string is %x2E ("."):

        Let cookie-domain be the attribute-value without the leading %x2E
        (".") character.

     Otherwise:

        Let cookie-domain be the entire attribute-value.
    }
    //if Domain[1] = '.' then Domain := Copy(Domain, 2, Length(Domain));
    
    if not IsDomainAcceptable(URLHost, Domain) then
    begin
      ParamList.Free;
      Exit;
    end;
  end;
  
  Path := ParamList.Values['path'];  
  
  Expires := ParamList.Values['expires'];
  if (Length(Expires) > 0) then ExpiresDate := CookieStrToDateTime(Expires);
  
  Result := True;
  
  ParamList.Free;
  
end;


function TCookie.IsExpired(Dt: TDateTime): Boolean;
begin
  Result := (ExpiresDate <> 0.0) and (ExpiresDate < (Dt - OffsetFromUTC));
end;


function TCookie.TheSame(C: TCookie): Boolean;
begin
  Result := (EffectiveDomain = C.EffectiveDomain) and (EffectivePath = C.EffectivePath) and (Name = C.Name) and (URLProto = C.URLProto);
end;


function TCookie.AsString: string;
begin
  Result := URLProto + ';' + URLHost + ';' + Name + ';' + Value + ';' + Expires + ';' + Path+ ';' + Domain;
end;


function TCookie.GetStringForSave: string;
begin
  Result := URLProto + '|' + URLHost + '|' + URLPath + '|' + Double2Str(ExpiresDate) + '|' + 
            Name + '|' + Value + '|' + Expires + '|' + Domain + '|' + Path;  
end;


function TCookie.SetCookieFromSaveStr(S: string): Boolean;
begin
  Result := False;
  URLProto := Parse(S, '|');
  URLHost := Parse(S, '|');
  URLPath := Parse(S, '|');
  ExpiresDate := Str2Double(Parse(S, '|'));
  Name := Parse(S, '|');
  Value := Parse(S, '|');
  Expires := Parse(S, '|');
  Domain := Parse(S, '|');
  Path := S;
  if (Length(URLHost) = 0) or (Length(URLPath) = 0) or (Length(Name) = 0) then Exit; 
  Result := True;
end;

// ========== ThiCookieManager ========== //


constructor ThiCookieManager.Create;
begin
  inherited;
  FCookieList := NewList;
  InitializeCriticalSection(FCSList);
end;


destructor ThiCookieManager.Destroy;
begin
  EnterCriticalSection(FCSList);
  DeleteCriticalSection(FCSList);
  FreeObjects;
  FCookieList.Free;
  inherited;
end;

procedure ThiCookieManager.InternalAddCookies(URL, Cookies: string; Source: Integer);
var 
  List: PStrList;
  I: Integer;
  Cook: TCookie;
  URLParser: TURI;
  Changed: Boolean;
  S: string;
begin

  if (Length(Cookies) = 0) or (Length(URL) = 0) then Exit;

  List := NewStrList;
  List.Text := Cookies;
  if Source = 0 then
  begin
    Changed := False;
    URLParser := TURI.Create(URL);
    if Length(URLParser.Protocol) = 0 then URLParser.Protocol := 'http';
    
    Cook := TCookie.Create;
    for I := 0 to List.Count - 1 do
    begin
      if Cook.SetCookie(URLParser, List.Items[I]) then
      begin
        AddToList(Cook);
        Changed := True;
        Cook := TCookie.Create;
      end;
    end;
    Cook.Free;
    
    URLParser.Free;
    if Changed then _hi_OnEvent(_event_onChange);
  end
  else
  begin
    for I := 0 to List.Count - 1 do
    begin
      S := List.Items[I];
      Replace(S, 'HttpOnly', ''); //  IE   HttpOnly,       
      Replace(S, 'httponly', '');
      InternetSetCookie(PChar(URL), nil, PChar(S));
    end;
  end;
  List.Free;
end;

function ThiCookieManager.InternalGetCookies(URL: string; Source: Integer): string;
var
  L: DWORD;
  Dt: TDateTime;
  I: Integer;
  Cook: TCookie;
  URLParser: TURI;
begin
  Result := '';
  if Length(URL) = 0 then Exit;

  Dt := Now;
  if Source = 0 then
  begin
    URLParser := TURI.Create(URL);
    if Length(URLParser.Protocol) = 0 then URLParser.Protocol := 'http';
    
    EnterCriticalSection(FCSList);
      for I := FCookieList.Count - 1 downto 0 do
      begin
        Cook := FCookieList.Items[I];
        if Cook.IsExpired(Dt) then 
        begin
          Cook.Free;
          FCookieList.Delete(I);
        end
        else
        begin
          if Cook.IsValidCookie(URLParser) then
          begin
            if Length(Result) = 0 then Result := 'Cookie: ' + Cook.ClientCookie
            else Result := Result + '; ' + Cook.ClientCookie;
          end;
        end;
      end;
    LeaveCriticalSection(FCSList);
    
    URLParser.Free;
  end
  else
  begin
    if InternetGetCookie(PChar(URL), nil, nil, L) and (L > 0)
    then
    begin
      SetLength(Result, L-1); //   NULL
      if InternetGetCookie(PChar(URL), nil, PChar(Result), L) then
        Result := 'Cookie: ' + Result
      else
      begin
        Result := '';
      end;
    end;
  end;
end;

procedure ThiCookieManager.AddCookies(URL, Cookies: string); //   
begin
  //_debug('ThiCookieManager.AddCookies: '+URL+#13#10+Cookies);
  InternalAddCookies(URL, Cookies, _prop_Source);
end;

function ThiCookieManager.GetCookies(URL: string): string;
begin
  Result := InternalGetCookies(URL, _prop_Source);
  //_debug('ThiCookieManager.GetCookies: '+URL+#13#10+Result);
end;

function ThiCookieManager.GetInterfaceCookieMan: ThiCookieManager;
begin
  Result := Self;
end;

procedure ThiCookieManager.FreeObjects;
var
  I: Integer;
begin
  for I := FCookieList.Count - 1 downto 0 do
    TCookie(FCookieList.Items[I]).Free;
end;


procedure ThiCookieManager.AddToList(C: TCookie);
var
  I: Integer;
  Cook: TCookie;
begin
  EnterCriticalSection(FCSList);
    for I := 0 to FCookieList.Count - 1 do
    begin
      Cook := FCookieList.Items[I];
      if Cook.TheSame(C) then
      begin
        Cook.Value := C.Value;
        Cook.Expires := C.Expires;
        Cook.ExpiresDate := C.ExpiresDate;
        C.Free;
        LeaveCriticalSection(FCSList);
        Exit;
      end;
    end;
    FCookieList.Add(C);
  LeaveCriticalSection(FCSList);
end;


procedure ThiCookieManager._work_doAddCookies;
var
  s: string;
begin
  s := ReadString (_Data, _data_Text);
  InternalAddCookies(ReadString (_Data, _data_URL), s, _prop_Source);
end;

procedure ThiCookieManager._work_doAddIECookies;
var
  s: string;
begin
  s := ReadString (_Data, _data_Text);
  InternalAddCookies(ReadString (_Data, _data_URL), s, 1);
end;


procedure ThiCookieManager._work_doGetCookies;
begin
  FResult := InternalGetCookies(ReadString (_Data, _data_URL), _prop_Source);

  if Length(FResult) > 0 then
  begin
    FResult := FResult + #13#10;
    _hi_CreateEvent(_Data, @_event_onGetCookies, FResult);
  end;
end;

procedure ThiCookieManager._work_doGetIECookies;
begin
  FResult := InternalGetCookies(ReadString (_Data, _data_URL), 1);

  if Length(FResult) > 0 then
  begin
    FResult := FResult + #13#10;
    _hi_CreateEvent(_Data, @_event_onGetCookies, FResult);
  end;
end;


procedure ThiCookieManager._work_doSave;
var
  f: string;
  List: PStrList;
  I: Integer;
  Cook: TCookie;
  Dt: TDateTime;
begin
  f := ReadString (_Data, _data_Filename, _prop_Filename);
  
  if Length(f) = 0 then exit;
  
  List := NewStrList;
  Dt := Now;
  
  EnterCriticalSection(FCSList);
    for I := 0 to FCookieList.Count - 1 do
    begin
      Cook := FCookieList.Items[I];
      if (Cook.ExpiresDate > 0) and not Cook.IsExpired(Dt) then 
      begin
        List.Add(Cook.GetStringForSave);
      end;
    end;
  LeaveCriticalSection(FCSList);
  
  List.SaveToFile(f);
  List.Free;
end;


procedure ThiCookieManager._work_doLoad;
var
  f: string;
  List: PStrList;
  I: Integer;
  Cook: TCookie;
  Dt: TDateTime;
begin
  f := ReadString (_Data, _data_Filename, _prop_Filename);
  
  if Length(f) = 0 then exit;
  
  Dt := Now;
  
  EnterCriticalSection(FCSList);
  
  FreeObjects;
  FCookieList.Clear;
  
  List := NewStrList;
  List.LoadFromFile(f);
  
  Cook := TCookie.Create;
  for I := 0 to List.Count - 1 do
  begin
    if Cook.SetCookieFromSaveStr(List.Items[I]) and not Cook.IsExpired(Dt) then
    begin
      FCookieList.Add(Cook);
      Cook := TCookie.Create;
    end;
  end;
  
  LeaveCriticalSection(FCSList);
  
  Cook.Free;
  List.Free;
  _hi_CreateEvent(_Data, @_event_onChange);
end;


procedure ThiCookieManager._work_doDelete;
var
  ind: Integer;
begin
  ind := ReadInteger(_Data, _data_Index, 0);
  EnterCriticalSection(FCSList);
    if (ind >= 0) and (ind < FCookieList.Count) then
    begin
      TCookie(FCookieList.Items[ind]).Free;
      FCookieList.Delete(ind);
      _hi_CreateEvent(_Data, @_event_onChange);
    end;
  LeaveCriticalSection(FCSList);
end;


procedure ThiCookieManager._work_doValue;
var
  ind: Integer;
  val: string;
begin
  ind := ReadInteger(_Data, _data_Index, 0);
  val := ToString(_Data);
  EnterCriticalSection(FCSList);
    if (ind >= 0) and (ind < FCookieList.Count) then
    begin
      TCookie(FCookieList.Items[ind]).Value := val;
      _hi_CreateEvent(_Data, @_event_onChange);
    end;
  LeaveCriticalSection(FCSList);
end;


procedure ThiCookieManager._work_doClear;
begin
  EnterCriticalSection(FCSList);
    FreeObjects;
    FCookieList.Clear;
  LeaveCriticalSection(FCSList);
  _hi_CreateEvent(_Data, @_event_onChange);
end;


procedure ThiCookieManager._var_Cookies;
begin
  dtString(_Data, FResult);
end;


// ================================================= //

function ThiCookieManager._Get;
var
  ind:Integer;
begin
  ind := ToIntIndex(Item);
  EnterCriticalSection(FCSList);
    if (ind >= 0) and (ind < FCookieList.Count) then
    begin
      Result := True;
      dtString(Val, TCookie(FCookieList.Items[ind]).AsString);
    end
    else
      Result := False;
  LeaveCriticalSection(FCSList);
end;


function ThiCookieManager._Count;
begin
  Result := FCookieList.Count;
end;


procedure ThiCookieManager._var_List;
begin
  if FArr = nil then FArr := CreateArray(nil, _Get, _Count, nil);
  dtArray(_Data, FArr);
end;

// ================================================= //


initialization
  OffsetFromUTC := GetOffsetFromUTC;

end.