unit hiHTTPClient;

interface

{$I share.inc}

uses Kol, Share, WinInet, WinHTTP, Windows, Debug, uri, hiCookieManager;

type

  ThiHTTPClient = class(TDebug)
    private
      FHSession: HINTERNET;
      FHConnect: HINTERNET;
      FHRequest: HINTERNET;
      FAbort: Boolean;
      FRedirected: Boolean;
      
      FURLParser: TURI;
      
      FStatusHeader: Integer;
      FContLenHeader: Integer;
      FBodyLength: Integer;
      FHasBody: Boolean;
      
      FContTypeHeader: string;
      FTmpBody: string;
      FMethod: WideString;
      
      FdtURL: string;
      FdtFileName: string;
      FdtHeaders: string;
      FdtHTTPUser: string;
      FdtHTTPPass: string;
      FdtPostData: string;
      FdtProxy: string;
      FdtProxyUser: string;
      FdtProxyPass: string;
      FdtReferer: string;
      FdtUploadFile: string;

      procedure CancelRequest;
      
      procedure InternalDownload(var _Data: TData; Mode: Integer);
      procedure InternalLoadFile(var _Data: TData);
      procedure InternalLoadStream(var _Data: TData);
      procedure InternalLoadString(var _Data: TData);
      procedure ReadPoints(var _Data: TData; ReadFileName: Boolean);
      function RetrieveData(var _Data: TData; Stream: PStream): Boolean;
      procedure PrepareBody(var _Data: TData);
      function SendRequest(var _Data: TData): Boolean;
      function WriteBody(var _Data: TData): Boolean;
      
    public
      _prop_Boundary: string;
      _prop_CustomMethod: string;
      _prop_FileField: string;
      _prop_FileName: string;
      _prop_Headers: string;
      _prop_PostData: string;
      _prop_Proxy: string;
      _prop_ProxyUser: string;
      _prop_ProxyPass: string;
      _prop_Referer: string;
      _prop_UploadFile: string;
      _prop_URL: string;
      _prop_UserAgent: string;
      _prop_HTTPUser: string;
      _prop_HTTPPass: string;
      _prop_AuthCheck: Integer;
      _prop_AutoRedirect: Integer;
      _prop_ConnectTimeout: Integer;
      _prop_EncType: Integer;
      _prop_Method: Integer;
      _prop_RecvTimeout: Integer;
      _prop_SendTimeout: Integer;
      _prop_UseSystemProxy: Integer;
      _prop_CookieManager: ThiCookieManager;

      _data_FileName: THI_Event;
      _data_HTTPPass: THI_Event;
      _data_HTTPUser: THI_Event;
      _data_Headers: THI_Event;
      _data_PostData: THI_Event;
      _data_Proxy: THI_Event;
      _data_ProxyPass: THI_Event;
      _data_ProxyUser: THI_Event;
      _data_Referer: THI_Event;
      _data_SaveTo: THI_Event;
      _data_Stream: THI_Event;
      _data_URL: THI_Event;
      _data_UploadFile: THI_Event;
      
      _event_onStateChange: THI_Event;
      _event_onDownloadProgress: THI_Event;
      _event_onError: THI_Event;
      _event_onFinish: THI_Event;
      _event_onHeadersAvail: THI_Event;
      _event_onRedirect: THI_Event;
      _event_onUploadProgress: THI_Event;
      
      constructor Create;
      destructor Destroy; override;
      
      procedure _work_doAbort(var _Data: TData; Index: Word);
      procedure _work_doBoundary(var _Data: TData; Index: Word);
      procedure _work_doCustomMethod(var _Data: TData; Index: Word);
      procedure _work_doFileField(var _Data: TData; Index: Word);
      procedure _work_doEncType(var _Data: TData; Index: Word);
      procedure _work_doExecute(var _Data: TData; Index: Word);
      procedure _work_doLoadFile(var _Data: TData; Index: Word);
      procedure _work_doLoadStream(var _Data: TData; Index: Word);
      procedure _work_doLoadString(var _Data: TData; Index: Word);
      procedure _work_doUseSystemProxy(var _Data: TData; Index: Word);
      
      procedure _var_Status(var _Data: TData; Index: Word);
      procedure _var_ContentLength(var _Data: TData; Index: Word);
      procedure _var_ContentType(var _Data: TData; Index: Word);
      procedure _var_CurrentURL(var _Data: TData; Index: Word);
      procedure _var_RawHeaders(var _Data: TData; Index: Word);
      procedure _var_StatusText(var _Data: TData; Index: Word);
  end;

implementation


const
  RequestMethods: array [0..2] of WideString = ('GET', 'POST', 'HEAD');
  
  AUTH_CHECK_DISABLED = 0;
  AUTH_CHECK_PROXY = 1;
  AUTH_CHECK_ALL = 2;
  
  METHOD_GET = 0;
  METHOD_POST = 1;
  METHOD_HEAD = 2;
  
  ENC_TYPE_MULTIPART = 0;
  ENC_TYPE_URL_ENC = 1;
  ENC_TYPE_NONE = 2;


//  Replace = True     
//  Replace = False       
procedure AddHeadersToRequest(RequestH: HINTERNET; const Headers: string; Replace: Boolean = True);
var
  F: DWORD;
begin
  if Headers = '' then Exit;
  
  F := WINHTTP_ADDREQ_FLAG_ADD;
  if Replace then F := F or WINHTTP_ADDREQ_FLAG_REPLACE;
  
  WinHttpAddRequestHeaders(
    RequestH,
    PWideChar(WideString(Headers)),
    Length(Headers),
    F
  );
end;

function ChooseAuthScheme(RequestH: HINTERNET): DWORD;
var
  SuppSchemes, FirstScheme, Target: DWORD;
begin
  Result := WINHTTP_AUTH_SCHEME_DIGEST;
  if WinHttpQueryAuthSchemes(RequestH, SuppSchemes, FirstScheme, Target) then
  begin
    if SuppSchemes and WINHTTP_AUTH_SCHEME_DIGEST <> 0 then
      Result := WINHTTP_AUTH_SCHEME_DIGEST
    else
      Result := WINHTTP_AUTH_SCHEME_BASIC;
  end;
end;

function GetHTTPOptionDWORD(Handle: HINTERNET; Option: DWORD): DWORD;
var
  L: DWORD;
begin
  L := SizeOf(DWORD);
  WinHttpQueryOption(Handle, Option, @Result, L);
end;

function GetHTTPOptionString(Handle: HINTERNET; Option: DWORD): string;
var
  Buffer: WideString;
  BufLen: DWORD;
  Rslt: Boolean;
label
  start;
const
  UrlBufSize = 250;
begin
  Result := '';
  if Handle = nil then Exit;
  
  BufLen := UrlBufSize;
  SetLength(Buffer, UrlBufSize);
  
start:
  Rslt := WinHttpQueryOption(Handle, Option, PWideChar(Buffer), BufLen);

  if (Rslt = False) and (GetLastError() = ERROR_INSUFFICIENT_BUFFER) then
  begin
    SetLength(Buffer, BufLen div SizeOf(WideChar));
    goto start;
  end;

  if Rslt then
  begin
    SetLength(Buffer, BufLen div SizeOf(WideChar));
    Result := Buffer;
  end;

end;

function ReadResponseHeader(Handle: HINTERNET; HeaderIdx: DWORD; PosIdx: PDWORD = nil): string; overload;
var
  Buffer: WideString;
  BufLen, PrevIdx: DWORD;
  Rslt: Boolean;
label
  start;
const
  HeaderBufSize = 250;
begin
  Result := '';
  PrevIdx := 0;
  if Handle = nil then Exit;
  
  BufLen := HeaderBufSize;
  SetLength(Buffer, HeaderBufSize);
  
start:
  if PosIdx <> nil then PrevIdx := PosIdx^;
  Rslt := WinHttpQueryHeaders(
             Handle,
             HeaderIdx,
             nil,
             PWideChar(Buffer),
             BufLen,
             PosIdx
           );

  if (Rslt = False) and (GetLastError() = ERROR_INSUFFICIENT_BUFFER) then
  begin
    SetLength(Buffer, BufLen div SizeOf(WideChar));
    if PosIdx <> nil then PosIdx^ := PrevIdx;
    goto start;
  end;


  if Rslt then
  begin
    SetLength(Buffer, BufLen div SizeOf(WideChar));
    Result := Buffer;
  end;
    
end;

function ReadResponseHeader(Handle: HINTERNET; HeaderName: string; PosIdx: PDWORD = nil): string; overload;
var
  Buffer: WideString;
  BufLen, PrevIdx: DWORD;
  Rslt: Boolean;
label
  start;
const
  HeaderBufSize = 250;
begin
  Result := '';
  PrevIdx := 0;
  if Handle = nil then Exit;
  
  BufLen := HeaderBufSize;
  SetLength(Buffer, HeaderBufSize);

start:
  if PosIdx <> nil then PrevIdx := PosIdx^;
  Rslt := WinHttpQueryHeaders(
             Handle,
             WINHTTP_QUERY_CUSTOM,
             PWideChar(WideString(HeaderName)),
             PWideChar(Buffer),
             BufLen,
             PosIdx //WINHTTP_NO_HEADER_INDEX
           );

  if (Rslt = False) and (GetLastError() = ERROR_INSUFFICIENT_BUFFER) then
  begin
    SetLength(Buffer, BufLen div SizeOf(WideChar));
    if PosIdx <> nil then PosIdx^ := PrevIdx;
    goto start;
  end;

  if Rslt then
  begin
    SetLength(Buffer, BufLen div SizeOf(WideChar));
    Result := Buffer;
  end;
    
end;

function ReadResponseCookies(Handle: HINTERNET): string;
var
  I: DWORD;
  Lst: PStrList;
  S: string;
begin
  Result := '';
  I := 0;
  Lst := NewStrList;
  
  while True do
  begin
    S := ReadResponseHeader(Handle, WINHTTP_QUERY_SET_COOKIE, @I);
    if S <> '' then Lst.Add(S)
    else Break;
  end;
  if Lst.Count > 0 then Result := Lst.Text;
  Lst.Free;
end;

procedure SetAuthCredentials(RequestH: HINTERNET; const User, Pass: string; Target, Scheme: DWORD);
begin
  if (User = '') or (Pass = '') then Exit;

  WinHttpSetCredentials(
    RequestH,
    Target,
    Scheme,
    PWideChar(WideString(User)),
    PWideChar(WideString(Pass)),
    nil
  );
end;

procedure SetHTTPOptionDWORD(RequestH: HINTERNET; Option, Value: DWORD);
begin
  WinHttpSetOption(RequestH, Option, @Value, SizeOf(DWORD));
end;








const
  CallbackFlags = 
    WINHTTP_CALLBACK_FLAG_RESOLVE_NAME or
    WINHTTP_CALLBACK_FLAG_CONNECT_TO_SERVER or
    WINHTTP_CALLBACK_FLAG_SEND_REQUEST or
    WINHTTP_CALLBACK_FLAG_RECEIVE_RESPONSE or
    WINHTTP_CALLBACK_FLAG_REDIRECT;

procedure StatusCallback(Handle: HINTERNET; dwContext: Pointer; dwInternetStatus: DWORD; lpvStatusInformation: Pointer; dwStatusInformationLength: DWORD); stdcall;
var
  U: string;
  UrlP: PWideChar;
  Client: ThiHTTPClient;
begin

  Client := dwContext;
  if dwInternetStatus = WINHTTP_CALLBACK_STATUS_REDIRECT then
  begin
    UrlP := lpvStatusInformation;
    U := UrlP;
    Client.FRedirected := True;
      
    _hi_OnEvent(Client._event_onRedirect, U);
    

    //  ,    
    AddHeadersToRequest(Handle, 'Cookie: ');
    
    if Client._prop_CookieManager <> nil then
    begin
      //       
      Client._prop_CookieManager.AddCookies(GetHTTPOptionString(Handle, WINHTTP_OPTION_URL), ReadResponseCookies(Handle));

      //     
      U := Client._prop_CookieManager.GetCookies(U);
      if U <> '' then AddHeadersToRequest(Handle, U);
    end;
    Exit;
  end;
  
  _hi_OnEvent(Client._event_onStateChange, Trunc(ln(dwInternetStatus) / ln(2)));
end;






constructor ThiHTTPClient.Create;
begin
  inherited;
  FURLParser := TURI.Create;
end;

destructor ThiHTTPClient.Destroy;
begin
  CancelRequest;
  FURLParser.Free;
  inherited;
end;

procedure ThiHTTPClient.CancelRequest;
begin
  FAbort := True;

  if FHRequest <> nil then WinHttpCloseHandle(FHRequest);
  if FHConnect <> nil then WinHttpCloseHandle(FHConnect);
  if FHSession <> nil then WinHttpCloseHandle(FHSession);
  
  FHRequest := nil;
  FHConnect := nil;
  FHConnect := nil;
  
  FBodyLength := 0;
  FTmpBody := '';
end;

procedure ThiHTTPClient.InternalDownload(var _Data: TData; Mode: Integer);
begin
  ReadPoints(_Data, Mode = 2); //     doLoadFile,     
  if SendRequest(_Data) then
  begin
    _hi_OnEvent(_event_onHeadersAvail, FStatusHeader);
    if not FAbort then
    begin
      if Mode = -1 then
      begin
        Mode := ReadInteger(_Data, _data_SaveTo, 0);
        if Mode = 2 then FdtFileName := ReadString(_Data, _data_FileName, _prop_FileName);
      end;
      case Mode of
        1: InternalLoadString(_Data);
        2: InternalLoadFile(_Data);
        3: InternalLoadStream(_Data);
      else
        CancelRequest;
        _hi_CreateEvent(_Data, @_event_onFinish);
      end;
    end;
  end;
  
  FdtURL := '';
  FdtFileName := '';
  FdtHeaders := '';
  FdtHTTPUser := '';
  FdtHTTPPass := '';
  FdtPostData := '';
  FdtProxy := '';
  FdtProxyUser := '';
  FdtProxyPass := '';
  FdtReferer := '';
  FdtUploadFile := '';
end;

procedure ThiHTTPClient.InternalLoadFile(var _Data: TData);
var
  Fn: string;
  St: PStream;
begin
  if _prop_Method = METHOD_HEAD then
  begin
    CancelRequest;
    _hi_CreateEvent(_Data, @_event_onFinish);
    Exit;
  end;
  
  Fn := FdtFileName;

  St := NewWriteFileStream(Fn);
  
  if St.Handle = INVALID_HANDLE_VALUE then
  begin
    _hi_CreateEvent(_Data, @_event_onError, Integer(GetLastError));
    St.Free;
    CancelRequest;
    Exit;
  end;

  if RetrieveData(_Data, St) then _hi_CreateEvent(_Data, @_event_onFinish);
  
  St.Free;
end;

procedure ThiHTTPClient.InternalLoadStream(var _Data: TData);
var
  St: PStream;
begin
  if _prop_Method = METHOD_HEAD then
  begin
    CancelRequest;
    _hi_CreateEvent(_Data, @_event_onFinish);
    Exit;
  end;
  
  St := ReadStream(_Data, _data_Stream);
  if St = nil then
  begin
    CancelRequest;
    Exit;
  end;
  if RetrieveData(_Data, St) then _hi_CreateEvent(_Data, @_event_onFinish);
end;

procedure ThiHTTPClient.InternalLoadString(var _Data: TData);
var
  St: PStream;
  S: string;
begin
  if _prop_Method = METHOD_HEAD then
  begin
    CancelRequest;
    _hi_CreateEvent(_Data, @_event_onFinish);
    Exit;
  end;
  
  //TODO:  .    
  St := NewMemoryStream;
  
  if RetrieveData(_Data, St) then
  begin
    SetLength(S, St.Size);
    Move(St.Memory^, S[1], St.Size);
    _hi_CreateEvent(_Data, @_event_onFinish, S);
  end;
  
  St.Free;
end;

procedure ThiHTTPClient.ReadPoints(var _Data: TData; ReadFileName: Boolean); //   
begin
  FdtURL := ReadString(_Data, _data_URL, _prop_URL);
  if ReadFileName then FdtFileName := ReadString(_Data, _data_FileName, _prop_FileName);
  FdtHeaders := ReadString(_Data, _data_Headers, _prop_Headers);
  FdtHTTPUser := ReadString(_Data, _data_HTTPUser, _prop_HTTPUser);
  FdtHTTPPass := ReadString(_Data, _data_HTTPPass, _prop_HTTPPass);
  
  if _prop_CustomMethod <> '' then
  begin
    FMethod := _prop_CustomMethod;
    FHasBody := True;
  end
  else
  begin
    FMethod := RequestMethods[_prop_Method];
    FHasBody := (_prop_Method = METHOD_POST);
  end;
  if FHasBody then FdtPostData := ReadString(_Data, _data_PostData, _prop_PostData);
  
  if _prop_UseSystemProxy = 0 then FdtProxy := ReadString(_Data, _data_Proxy, _prop_Proxy);
  FdtProxyUser := ReadString(_Data, _data_ProxyUser, _prop_ProxyUser);
  FdtProxyPass := ReadString(_Data, _data_ProxyPass, _prop_ProxyPass);
  FdtReferer := ReadString(_Data, _data_Referer, _prop_Referer);
  FdtUploadFile := ReadString(_Data, _data_UploadFile, _prop_UploadFile);
end;

function ThiHTTPClient.RetrieveData(var _Data: TData; Stream: PStream): Boolean;
var
  Buf: string;
  BytesRead: Cardinal;
  Total: Integer;
const
  BUFFER_SIZE = 10*1024;
begin
  Result := False;
  Total := 0;
  SetLength(Buf, BUFFER_SIZE);
  
  while not FAbort do
  begin
    if not WinHttpReadData(FHRequest, @Buf[1], BUFFER_SIZE, BytesRead) then
    begin
      if not FAbort then
      begin
        _hi_CreateEvent(_Data, @_event_onError, Integer(GetLastError));
        CancelRequest;
      end;
      Exit;
    end;
    
    if FAbort then Exit;
    
    if BytesRead = 0 then //  
    begin
      Result := True;
      CancelRequest;
      Exit;
    end;
    
    if Stream.Write(Buf[1], BytesRead) < BytesRead then
    begin
      _hi_CreateEvent(_Data, @_event_onError, Integer(GetLastError));
      CancelRequest;
      Exit;
    end;
    
    Inc(Total, BytesRead);
    _hi_OnEvent(_event_onDownloadProgress, Total);
  end;
  
end;

procedure ThiHTTPClient.PrepareBody(var _Data: TData);
var
  Key, Val, Dt: string;
  FS: PStream;
  Body, PostDt: PStrList;
  I, L: Integer;
begin
  if _prop_Boundary = '' then _prop_Boundary := '----HiHtTpClIeNt';
  
  Dt := FdtPostData;
  
  if _prop_EncType = ENC_TYPE_MULTIPART then
  begin
    Body := NewStrList;
    if Dt <> '' then
    begin
      PostDt := NewStrList;
      PostDt.Text := Dt;
      for I := 0 to PostDt.Count - 1 do
      begin
        Val := PostDt.Items[I];
        Key := Parse(Val, '=');
        if Key <> '' then
        begin
          Body.Add('--' + _prop_Boundary);
          Body.Add('Content-Disposition: form-data; name="' + Key + '"');
          Body.Add('');
          Body.Add(Val);
        end;
      end;
      PostDt.Free;
    end;
    
    L := 0;
    
    if (_prop_FileField <> '') and (FdtUploadFile <> '') then
    begin
      FS := NewReadFileStream(FdtUploadFile);
        if FS.Handle <> INVALID_HANDLE_VALUE then
        begin
          Body.Add('--' + _prop_Boundary);
          Body.Add('Content-Disposition: form-data; name="' + _prop_FileField + '"; filename="' + ExtractFileName(FdtUploadFile) + '"');
          Body.Add('Content-Type: application/octet-stream');
          Body.Add('');
          Body.Add(''); //  #13#10   
          L := FS.Size;
        end
        else // TODO:  ,    ?
          FdtUploadFile := '';
      FS.Free;
    end;

    FTmpBody := Body.Text;
    //   #13#10 (    Boundary      )
    FTmpBody := Copy(FTmpBody, 1, Length(FTmpBody) - 2);
    
    //   \20\20   
    //     \r\n    POSTData  EncType=multipart/form-data
    //   "="  ""  "\r\n"  "\20\20"
    Replace(FTmpBody, #20#20, #13#10);
    
    Body.Free;
    FBodyLength := Length(FTmpBody) + L;
    
    if FBodyLength > 0 then
    begin
      FBodyLength := FBodyLength + Length(_prop_Boundary) + 8; // +    #13#10'--'+_prop_Boundary+'--'#13#10
      AddHeadersToRequest(FHRequest, 'Content-Type: multipart/form-data; boundary=' + _prop_Boundary);
      AddHeadersToRequest(FHRequest, 'Content-Length: '+Int2Str(FBodyLength));
    end;
  end
  else
  begin
    if Dt = '' then Exit;
    if _prop_EncType = ENC_TYPE_URL_ENC then
    begin
      Replace(Dt, #13#10, '&');
      AddHeadersToRequest(FHRequest, 'Content-Type: application/x-www-form-urlencoded');
    end;
    FTmpBody := Dt;
    FBodyLength := Length(FTmpBody);
    AddHeadersToRequest(FHRequest, 'Content-Length: '+Int2Str(FBodyLength));
  end;
end;

function ThiHTTPClient.SendRequest(var _Data: TData): Boolean;
var
  N: Cardinal;
  WStrP1, WStrP2: PWideChar;

  ProxyAuthSended, CanSendData, DataSended: Boolean;
  CurrUrl, LastAuthUrl: string;
  
label
  send_again;
begin
  CancelRequest;
  Result := False;
  FAbort := False;
  
  FStatusHeader := 0;
  FContLenHeader := 0;
  FContTypeHeader := '';

  
  FURLParser.URI := FdtURL;
  if FURLParser.Protocol <> 'https' then FURLParser.Protocol := 'http';
  
  
  N := WINHTTP_ACCESS_TYPE_NO_PROXY;
  WStrP1 := nil;
  WStrP2 := nil;
  
  
  if _prop_UseSystemProxy = 0 then
  begin
    if FdtProxy <> '' then
    begin
      N := WINHTTP_ACCESS_TYPE_NAMED_PROXY; 
      WStrP1 := PWideChar(WideString(FdtProxy));
      WStrP2 := PWideChar(WideString('<local>'));
    end;
  end
  else
    N := WINHTTP_ACCESS_TYPE_DEFAULT_PROXY;
  
  // --------- WinHttpOpen --------- //
  
  FHSession := WinHttpOpen(
    PWideChar(WideString(_prop_UserAgent)),
    N,
    WStrP1,
    WStrP2,
    0
  );

  if FHSession = nil then
  begin
    _hi_CreateEvent(_Data, @_event_onError, Integer(GetLastError));
    Exit;
  end;
  
  //SetHTTPOptionDWORD(FHSession, WINHTTP_OPTION_CONNECT_TIMEOUT, _prop_ConnectTimeout);
  //SetHTTPOptionDWORD(FHSession, WINHTTP_OPTION_SEND_TIMEOUT, _prop_SendTimeout);
  //SetHTTPOptionDWORD(FHSession, WINHTTP_OPTION_RECEIVE_TIMEOUT, _prop_RecvTimeout);
  WinHttpSetTimeouts(FHSession, 0, _prop_ConnectTimeout, _prop_SendTimeout, _prop_RecvTimeout);
  
  // -------------------------------- //


  //       -  
  if FURLParser.Port = '' then
  begin
    if FURLParser.Protocol = 'https' then N := 443 else N := 80;
  end
  else
    N := Str2Int(FURLParser.Port);


  // --------- WinHttpConnect --------- //
  
  FHConnect := WinHttpConnect(
    FHSession,
    PWideChar(WideString(FURLParser.Host)),
    N,
    0
  );

  if FHConnect = nil then
  begin
    _hi_CreateEvent(_Data, @_event_onError, Integer(GetLastError));
    CancelRequest;
    Exit;
  end;
  
  // ----------------------------------- //

  
  if FdtReferer <> '' then WStrP1 := PWideChar(WideString(FdtReferer)) else WStrP1 := nil;
  
  
  N := WINHTTP_FLAG_REFRESH;
  if FURLParser.Protocol = 'https' then N := N or WINHTTP_FLAG_SECURE;
  
  
  // --------- WinHttpOpenRequest --------- //
  
  FHRequest := WinHttpOpenRequest(
    FHConnect,
    PWideChar(FMethod), //pwszVerb
    PWideChar(WideString(FURLParser.GetFullURI([ofParams]))), //pwszObjectName
    'HTTP/1.1', //pwszVersion
    WStrP1, //pwszReferrer
    WINHTTP_DEFAULT_ACCEPT_TYPES, //ppwszAcceptTypes //
    N
  );

  if FHRequest = nil then
  begin
    _hi_CreateEvent(_Data, @_event_onError, Integer(GetLastError));
    CancelRequest;
    Exit;
  end;

  // ----------------------------------- //
  
  
  CurrUrl := GetHTTPOptionString(FHRequest, WINHTTP_OPTION_URL); // ,    

  // Callback 
  WinHttpSetStatusCallback(FHRequest, @StatusCallback, CallbackFlags, 0);
  
  //  
  if _prop_AutoRedirect = 0 then
    SetHTTPOptionDWORD(FHRequest, WINHTTP_OPTION_DISABLE_FEATURE, WINHTTP_DISABLE_REDIRECTS);


  FRedirected := False;
  DataSended := False;
  ProxyAuthSended := False;
  LastAuthUrl := '';
  
  CanSendData := (_prop_AuthCheck = AUTH_CHECK_DISABLED);
  
send_again:
  
  FBodyLength := 0;
   
  // ------    ------ //

  //   
  if _prop_CookieManager <> nil then
    AddHeadersToRequest(FHRequest, _prop_CookieManager.GetCookies(CurrUrl));
  
  if not FRedirected then //     
  begin
    if CanSendData then
    begin
      //  
      AddHeadersToRequest(FHRequest, FdtHeaders);

      //   POST
      if FHasBody then PrepareBody(_Data);
    end;
  end;

  // ----------------------------------- //


  
  // --------- WinHttpSendRequest --------- //
  
  if not WinHttpSendRequest(
           FHRequest,
           nil, //pwszHeaders
           0, //dwHeadersLength
           nil, //lpOptional
           0, //dwOptionalLength
           FBodyLength, //dwTotalLength
           Self
        )
  then
  begin
    _hi_CreateEvent(_Data, @_event_onError, Integer(GetLastError));
    CancelRequest;
    Exit;
  end;

  // ----------------------------------- //

  
  //   POST
  if FHasBody and CanSendData and (not FRedirected) then //     
    if not WriteBody(_Data) then Exit;


  // --------- WinHttpReceiveResponse --------- //
  
  if not WinHttpReceiveResponse(FHRequest, nil) then
  begin
    _hi_CreateEvent(_Data, @_event_onError, Integer(GetLastError));
    CancelRequest;
    Exit;
  end;

  // ----------------------------------- //

  CurrUrl := GetHTTPOptionString(FHRequest, WINHTTP_OPTION_URL); // ,     
  
  FStatusHeader := Str2Int(ReadResponseHeader(FHRequest, WINHTTP_QUERY_STATUS_CODE));
  FContLenHeader := Str2Int(ReadResponseHeader(FHRequest, WINHTTP_QUERY_CONTENT_LENGTH));
  FContTypeHeader := ReadResponseHeader(FHRequest, WINHTTP_QUERY_CONTENT_TYPE);
  

  if _prop_CookieManager <> nil then
  begin
    _prop_CookieManager.AddCookies(CurrUrl, ReadResponseCookies(FHRequest));
  end;
  
  // Proxy 
  if (FStatusHeader = 407) and (not ProxyAuthSended) then // TODO:   -    
  begin
    ProxyAuthSended := True;
    SetAuthCredentials(FHRequest, 
      FdtProxyUser, 
      FdtProxyPass, 
      WINHTTP_AUTH_TARGET_PROXY,
      ChooseAuthScheme(FHRequest)
    );
    CanSendData := _prop_AuthCheck <> AUTH_CHECK_ALL;
    DataSended := CanSendData;
    goto send_again;
  end;

  // HTTP 
  if (FStatusHeader = 401) and (LastAuthUrl <> CurrUrl) then //       URL      -  /
  begin
    LastAuthUrl := CurrUrl;

    SetAuthCredentials(FHRequest, 
      FdtHTTPUser, 
      FdtHTTPPass, 
      WINHTTP_AUTH_TARGET_SERVER,
      ChooseAuthScheme(FHRequest)
    );
    CanSendData := True;
    DataSended := True;
    goto send_again;
  end;
  
  if (_prop_AuthCheck <> AUTH_CHECK_DISABLED) and (FStatusHeader = 200) and (not FRedirected) and (not DataSended) then
  begin
    CanSendData := True;
    DataSended := True;
    goto send_again;
  end;
  
  Result := True;
 
end;

function ThiHTTPClient.WriteBody(var _Data: TData): Boolean;
var
  FS: PStream;
  I, R, Sz: Integer;
  L: DWORD;
  Buf: string;
  FinalBound: string;
const
  READ_BUFFER_SIZE = 8*1024;
begin
  Result := True;
  
  if (FBodyLength = 0) or (FTmpBody = '') then Exit;
  
  I := 0;
  Sz := Length(FTmpBody);
  while (I < Sz) and (not FAbort) do
  begin
    R := Sz - I;
    if R > READ_BUFFER_SIZE then R := READ_BUFFER_SIZE;
    
    if not WinHttpWriteData(FHRequest, @FTmpBody[I + 1], R, L) then
    begin
      if not FAbort then
      begin
        _hi_CreateEvent(_Data, @_event_onError, Integer(GetLastError));
        CancelRequest;
      end;
      Result := False;
      Exit;
    end;
    Inc(I, R);
  end;
  
  if FdtUploadFile <> '' then
  begin
    FS := NewReadFileStream(FdtUploadFile);
    if FS.Handle <> INVALID_HANDLE_VALUE then
    begin
      SetLength(Buf, READ_BUFFER_SIZE);
      Sz := FS.Size;
      I := 0;
      while (I < Sz) and (not FAbort) do
      begin
        R := FS.Read(Buf[1], READ_BUFFER_SIZE);
        if not WinHttpWriteData(FHRequest, @Buf[1], R, L) then
        begin
          if not FAbort then
          begin
            _hi_CreateEvent(_Data, @_event_onError, Integer(GetLastError));
            CancelRequest;
          end;
          Result := False;
          Break;
        end;
        Inc(I, R);
        _hi_OnEvent(_event_onUploadProgress, I);
        if R < READ_BUFFER_SIZE then Break;
      end;
    end
    else
    begin
      _hi_CreateEvent(_Data, @_event_onError, Integer(GetLastError));
      CancelRequest;
      Result := False;
    end;
    FS.Free;
  end;
  if (_prop_EncType = ENC_TYPE_MULTIPART) and (not FAbort) then
  begin
    FinalBound := #13#10'--' + _prop_Boundary + '--'#13#10;
    WinHttpWriteData(FHRequest, Pointer(FinalBound), Length(FinalBound), L);
  end;
end;

procedure ThiHTTPClient._work_doAbort(var _Data: TData; Index: Word);
begin
  CancelRequest;
end;

procedure ThiHTTPClient._work_doBoundary(var _Data: TData; Index: Word);
begin
  _prop_Boundary := ToString(_Data);
end;

procedure ThiHTTPClient._work_doCustomMethod(var _Data: TData; Index: Word);
begin
  _prop_CustomMethod := ToString(_Data);
end;

procedure ThiHTTPClient._work_doFileField(var _Data: TData; Index: Word);
begin
  _prop_FileField := ToString(_Data);
end;

procedure ThiHTTPClient._work_doEncType(var _Data: TData; Index: Word);
begin
  _prop_EncType := ToInteger(_Data);
end;

procedure ThiHTTPClient._work_doExecute(var _Data: TData; Index: Word);
begin
  InternalDownload(_Data, -1);
end;

procedure ThiHTTPClient._work_doLoadFile(var _Data: TData; Index: Word);
begin
  InternalDownload(_Data, 2);
end;

procedure ThiHTTPClient._work_doLoadStream(var _Data: TData; Index: Word);
begin
  InternalDownload(_Data, 3);
end;

procedure ThiHTTPClient._work_doLoadString(var _Data: TData; Index: Word);
begin
  InternalDownload(_Data, 1);
end;

procedure ThiHTTPClient._work_doUseSystemProxy(var _Data: TData; Index: Word);
begin
  _prop_UseSystemProxy := ToInteger(_Data);
end;



procedure ThiHTTPClient._var_Status(var _Data: TData; Index: Word);
begin
  dtInteger(_Data, FStatusHeader);
end;

procedure ThiHTTPClient._var_ContentLength(var _Data: TData; Index: Word);
begin
  dtInteger(_Data, FContLenHeader);
end;

procedure ThiHTTPClient._var_ContentType(var _Data: TData; Index: Word);
begin
  dtString(_Data, FContTypeHeader);
end;

procedure ThiHTTPClient._var_CurrentURL(var _Data: TData; Index: Word);
begin
  dtString(_Data, GetHTTPOptionString(FHRequest, WINHTTP_OPTION_URL));
end;

procedure ThiHTTPClient._var_RawHeaders(var _Data: TData; Index: Word);
begin
  dtString(_Data, ReadResponseHeader(FHRequest, HTTP_QUERY_RAW_HEADERS_CRLF));
end;

procedure ThiHTTPClient._var_StatusText(var _Data: TData; Index: Word);
begin
  dtString(_Data, ReadResponseHeader(FHRequest, HTTP_QUERY_STATUS_TEXT));
end;


end.
